/*
 * Peter Saunders pjob.c
 * Copyright (C) 2000-2005 Peter Saunders
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <poll.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/resource.h>

#define DEFNUMTHREAD 10
#define DEFTIMEOUT 60

/* Structure for the process to be passed to the executing thread */
struct _process_t{
	gchar *exec;
	gchar *jobname;
	GTimer *timer;
	GPid pid;
	gint stat_loc;
	gchar *file_stdout;
	gchar *file_stderr;

	GError *err;
};

typedef struct _process_t process_t;

/* Globals for config setup */
static gint numthreads = DEFNUMTHREAD;
static gboolean verbose = FALSE;
static gboolean quiet = FALSE;
static gint timeout = DEFTIMEOUT;
static gchar *parseformat = NULL;
static gchar *jobparsename = NULL;
static gchar *command = NULL;
static gchar *outdir = NULL;
static gchar *infile = NULL;
static gchar *arglist = NULL;

/* Command line options */
static GOptionEntry options[] =
{
	{ "jobs", 'j', 0, G_OPTION_ARG_INT, &numthreads, "Number of jobs to run in parallel [DEFNUMTHREAD]", NULL },
	{ "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose, "Be verbose", NULL },
	{ "quiet", 'q', 0, G_OPTION_ARG_NONE, &quiet, "Do not print the output of the commands as they are running", NULL },
	{ "timeout", 't', 0, G_OPTION_ARG_INT, &timeout, "Timeout before process is killed [DEFTIMEOUT]", NULL },
	{ "format", 'f', 0, G_OPTION_ARG_STRING, &parseformat, "The order of substitiution to be used in the command", NULL },
	{ "jobname", 'n', 0, G_OPTION_ARG_STRING, &jobparsename, "If format is used, which variable to use for a output dir", NULL },
	{ "command", 'c', 0, G_OPTION_ARG_STRING, &command, "The command to be executed", NULL },
	{ "output", 'o', 0, G_OPTION_ARG_STRING, &outdir, "Directory to put all output into", NULL },
	{ "stdin", 'i', 0, G_OPTION_ARG_FILENAME, &infile, "Pass contents of filename into stdin of the executing process", NULL },
	{ "argsfile", 'a', 0, G_OPTION_ARG_FILENAME, &arglist, "File for list of argumenst if you dont want to use the command line", NULL },
	{ NULL }
};

/* Linked list of process jobs */
GList *proclist = NULL;

/* Take a process_t and execute it, and doing the "right thing" with the output */
void process_child(gpointer data, gpointer user_data){
	process_t *proc = (process_t*) data;
	GIOChannel *soutfile[2], *sout[2], *sinfile, *sin;
	gchar *execargv[4];
	int outpipes[2], inpipes[2];
	GError *err = NULL;

	struct pollfd fds[3];
	gint fdssize=2;

	proc->timer = g_timer_new();
	g_timer_start(proc->timer);

	/* Setup files in output dir if requested to do so */
	if(outdir != NULL){
		proc->file_stdout = g_strdup_printf("%s/%s-STDOUT", outdir, proc->jobname);
		soutfile[0] = g_io_channel_new_file(proc->file_stdout, "w", &err);

		if(soutfile[0] == NULL){
			g_printerr("Failed to open %s for writing: %s. Skipping job\n", proc->file_stdout, err->message);
			return;
		}
		
		proc->file_stderr = g_strdup_printf("%s/%s-STDERR", outdir, proc->jobname);
		soutfile[1] = g_io_channel_new_file(proc->file_stderr, "w", &err);

		if(soutfile[1] == NULL){
		        g_printerr("Failed to open %s for writing: %s. Skipping job\n", proc->file_stderr, err->message);
			return;
	        }       

	}


	/* Open stdin file to pass to the process */
	if(infile != NULL){
		sinfile = g_io_channel_new_file(infile, "r", NULL);
		pipe(inpipes);
	}

	/* Setup argv structure for job */
	if (verbose) g_fprintf(stderr, "Starting job '%s'\n", proc->jobname);
	execargv[0] = "/bin/sh";
	execargv[1] = "-c";
	execargv[2] = proc->exec;
	execargv[3] = NULL;

	/* Exec the job */
	if (infile == NULL){
		if( ! g_spawn_async_with_pipes(NULL, execargv, NULL, G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &(proc->pid), NULL, &(outpipes[0]), &(outpipes[1]), &err)){
			g_printerr("Failed to execute job %s: %s\n", proc->jobname, err->message);
			return;
		}
	}else{
		if( ! g_spawn_async_with_pipes(NULL, execargv, NULL, G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &(proc->pid), &(inpipes[1]), &(outpipes[0]), &(outpipes[1]), &err)){
			g_printerr("Failed to execute job %s: %s\n", proc->jobname, err->message);
			return;
		}
		close(inpipes[0]);
	}


	/* Make a stream out of the pipes for ease of reading from them */
	sout[0] = g_io_channel_unix_new(outpipes[0]);
	sout[1] = g_io_channel_unix_new(outpipes[1]);
	if(infile != NULL){
		sin = g_io_channel_unix_new(inpipes[1]);
		fds[2].fd = inpipes[1];
		fds[2].events = POLLOUT | POLLHUP;
		fdssize = 3;
	}


	/* Setup the poll events */
	fds[0].fd = outpipes[0];
	fds[1].fd = outpipes[1];
	fds[0].events = fds[1].events = POLLIN | POLLPRI | POLLHUP;
	fds[0].revents = fds[1].revents = fds[2].revents = 0;


	for(;;){
		gint x;
		gchar *readbuf;
		gint rdatasize, wdatasize;
		gboolean readdata = FALSE;
		fds[0].revents = fds[1].revents = fds[2].revents = 0;

		poll(fds, fdssize, -1);
		/* For stdout and stderr see if there is any data, and read it */
		for(x=0; x<2; x++){
			if((fds[x].revents & POLLIN) != 0){
				/* We have data to read */
				g_io_channel_read_line(sout[x], &readbuf, &rdatasize, NULL, NULL);
				if(rdatasize > 0){
					/* Print it if unless told not to */
					if(!quiet){
						g_printf("[%s] [%s] %s", proc->jobname, (x==0) ? "out" : "err", readbuf);
					}
					if(outdir != NULL){
						g_io_channel_write_chars(soutfile[x], readbuf, rdatasize, &wdatasize, NULL);
					}
					readdata = TRUE;
					free(readbuf);
				}

			}
		}
		/* See if we need to pump more data down stdin */
		if((fds[2].revents & POLLOUT) != 0){
			/* We have data we can write */
			gchar *nextline;
			gint nextlinesize;
			gint nextlinewritesize;
			GIOStatus s;

			/* Get the next line, and write it down the stream */
			s = g_io_channel_read_line(sinfile, &nextline, &nextlinesize, NULL, NULL);
			if (nextlinesize > 0){
				g_io_channel_write_chars(sin, nextline, nextlinesize, &nextlinewritesize, NULL);
			}
			if (s == G_IO_STATUS_EOF){
				g_io_channel_shutdown(sin, TRUE, NULL);
				sin = NULL;
				fdssize=2;
			}

		}
		/* Even if we did get a hangup - lets make sure there is no more data to read first by looping again */
		if (readdata) continue;

		if(((fds[0].revents & POLLHUP) != 0) && ((fds[1].revents & POLLHUP) != 0)) break;
	}	

	while((waitpid(proc->pid, &(proc->stat_loc), 0)) != proc->pid);

	g_timer_stop(proc->timer);

	/* If process exited cleanly */
	if (WIFEXITED(proc->stat_loc)){
		/* Get the exit code */
		if (verbose) g_fprintf(stderr, "Job '%s' exited with code %d. Exec time: %.2f\n", proc->jobname, WEXITSTATUS(proc->stat_loc), g_timer_elapsed(proc->timer, 0));
	}else{
		/* Otherwise - find out what it died with */
		/* TODO - this doesn't work quite right.. Mainly because its looking at the shell process, so the
		 * child of the /bin/sh which does get a signal, this isn't passed up. Although, it handly tells
		 * us if the /bin/sh gets a SEGV etc ;)
		 */
		g_fprintf(stderr, "Job %s exited with signal %d. Exec time: %.2f\n", proc->jobname, (WTERMSIG(proc->stat_loc)), g_timer_elapsed(proc->timer, 0));
	}


	g_io_channel_shutdown(sout[0], TRUE, NULL);
	g_io_channel_shutdown(sout[1], TRUE, NULL);

	if((infile != NULL) && (sin != NULL)){
		g_io_channel_shutdown(sin, TRUE, NULL);
	}
	if (outdir != NULL){
		g_io_channel_shutdown(soutfile[0], TRUE, NULL);
		g_io_channel_shutdown(soutfile[1], TRUE, NULL);
	}

	g_spawn_close_pid(proc->pid);


}

/* Takes a string str, a search string, find, and string to
 * replace all occurs of find with, replace. Returns a new
 * leaving original intact.
 */
gchar *strrep(gchar *str, gchar *find, gchar *replace){
	gchar *ptr, *oldptr;
	GString *newstr = g_string_new("");
	gssize len = strlen(str);
	gint findlen = strlen(find);

	ptr = g_strstr_len(str, len, find);
	oldptr=str;
	while(ptr != NULL){
		/* Copy in data up to this point */
		g_string_append_len (newstr, oldptr, (ptr - oldptr));
		/* Put in the replacement string */
		g_string_append(newstr, replace);

		oldptr = ptr + findlen;
		/* BUG - len will now be wrong. But, i only wanted a strstr anyway :) */
		ptr = g_strstr_len(oldptr, len, find);
	}

	/* Copy remains */
	g_string_append_len (newstr, oldptr, (ptr - oldptr));

	ptr = g_string_free(newstr, FALSE);

	return ptr;
}


/* Takes a cmd before substitution, takes the characters to be substituted
 * and a line for doign the substitution with. Fills in jobname
 */
gchar *genexeccmd(gchar *cmd, gchar *fmt, gchar *line, gchar **jobname){
	gchar *newexec, *ptr;
	int x;
	gchar *linesep = " ";
	gchar *fmtsep = " ";

	gchar **line_array;
	gchar **fmt_array;

	if ( fmt == NULL ){
		/* No format given - we'll just append the options to the end of the command */
		if(jobname != NULL) *jobname = g_strdup(line);
		return g_strdup_printf("%s %s", cmd, line);
	}

	line_array = g_strsplit(line, linesep, 0);
	fmt_array = g_strsplit(fmt, fmtsep, 0);

	if(jobparsename != NULL){
		if(jobname != NULL){
			for(x=0; fmt_array[x] != NULL; x++){
				if (line_array[x] == NULL) break;
				if((strcmp(fmt_array[x], jobparsename) == 0)){
					*jobname = g_strdup(line_array[x]);
					break;
				}
			}
		}
	}else{
		/* Not told us what they want.. We'll just use the first one */
		*jobname = g_strdup(line_array[0]);
	}

	newexec = g_strdup(cmd);
	for(x=0; line_array[x] != NULL; x++){
		if (fmt_array[x] == NULL) break;
		ptr = newexec;
		newexec = strrep(newexec, fmt_array[x], line_array[x]);
		free(ptr);
	}


	return newexec;
}

int main(int argc, char **argv){

	GThreadPool *procpool;
	GError *pp_err = NULL, *err = NULL;
	gint x;

	struct rlimit rlp;

	GOptionContext *optcontext;

	optcontext = g_option_context_new(" - parallel job executer");
	g_option_context_add_main_entries(optcontext, options, NULL);
	g_option_context_parse (optcontext, &argc, &argv, &err);

	if(command == NULL){
		g_printerr("Command required, see --help for more flags\n");
		exit(1);
	}

	if(verbose){
		g_printerr("Command '%s'\n", command);
		g_printerr("Timeout '%d'\n", timeout);
		g_printerr("Jobs '%d'\n", numthreads);
	}
	
	
	if(argc < 2 && arglist == NULL){
		/* We have no arguments */
		g_printerr("Missing arguments, see --help for details\n");
		exit(1);
	}

	if (!g_thread_supported ()){ 
		g_thread_init (NULL);
	}else{
		g_printerr("Threading not supported\n");
	}

	/* Up the number of FD's to the "hard" limit.
	 * This is mainly to get around the very small default
	 * solaris has, or 256
	 */
	getrlimit(RLIMIT_NOFILE, &rlp);
	rlp.rlim_cur = rlp.rlim_max;
	setrlimit(RLIMIT_NOFILE, &rlp);
	
	if(verbose) g_printerr("Creating a threadpool %d in size\n", numthreads);
	procpool = g_thread_pool_new(process_child, NULL, numthreads, FALSE, &pp_err);

	/* Generate the commands and push the job onto the thread pool */
	/* If no substituion is needed */
	if (arglist != NULL){
		GIOChannel *f;
		gchar *line;
		GIOStatus status;

		f = g_io_channel_new_file(arglist, "r", &err);
		if (f == NULL){
			g_printerr("Failed to open argfile: %s\n", err->message);
			exit(1);
		}
		status = g_io_channel_read_line(f, &line, NULL, NULL, &err);
		while(status==G_IO_STATUS_NORMAL){
			process_t *newproc = g_new(process_t, 1);
			newproc->err = NULL;
			
			line = g_strstrip(line);
		
			newproc->exec = genexeccmd(command, parseformat, line, &(newproc->jobname));
			proclist = g_list_append(proclist, (gpointer) newproc);
		
			if(verbose) g_printerr("Pushing command '%s' into thread pool queue\n", newproc->exec);
			g_thread_pool_push(procpool, (gpointer) newproc, &(newproc->err));
			
			status = g_io_channel_read_line(f, &line, NULL, NULL, &err);
		}
		g_io_channel_close(f);
		
	}else{
		/* substition is needed */
		for(x=1; x<argc; x++){
			process_t *newproc = g_new(process_t, 1);
			newproc->err = NULL;

			newproc->exec = genexeccmd(command, parseformat, argv[x],  &(newproc->jobname));

			proclist = g_list_append(proclist, (gpointer) newproc);

			if(verbose) g_printerr("Pushing command '%s' into thread pool queue\n", newproc->exec);
			g_thread_pool_push(procpool, (gpointer) newproc, &(newproc->err));
		}
	}
		

	/* Wait for the jobs to finish */
	/* TODO - Kill jobs that don't finish in time */
	while(g_thread_pool_get_num_threads(procpool) > 0){
		g_usleep(1000);
	}

	return 0;
}
